/**
 * @file
 * @brief The server thread code
 *
 *
 * Netsweeper Squid URL Redirector
 * This application is designed to work with the Netsweeper Internet Content Filtering System.
 *
 * Copyright 2010 Netsweeper Inc. All Rights Reserved.
 * Netsweeper, Netsweeper Inc. is either a copyrighted or registered trademark of Netsweeper Inc.
 */
#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <netdb.h>
#include <errno.h>
#include <string.h>
#include <pthread.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <syslog.h>

#include "simple_iplist.h"
#include "redir.h"
#include "protocol.h"
#include "request.h"
#include "log.h"
#include "utils.h"

#define TIMEOUT_STEP 1000

// Forward declarations

// connect to policy server
int st_connect(Connection* conn);
// send a request
int st_send(Connection* conn, Request* request);
// get a reply
int st_receive(Connection* conn, char* reply);

int st_connect_nointr(int socket, const struct sockaddr *addr, socklen_t addrlen);

/*
 * @brief continue the connect on EINTR
 */
int st_connect_nointr(int socket, const struct sockaddr *addr, socklen_t addrlen) {
	int n;
	for (;;) {
		n = connect(socket, addr, addrlen);
		if (n < 0 && errno == EINTR)
			continue;
		return n;
	}
}


/**
 * @brief The connection thread procedure with the following functionalities:
 * 	Set up and maintain connection to NSP server
 * 	Get requests from global queue
 * 	Send request to NSP
 * 	Receive reply from NSP
 * 	Write reply back to Squid
 *
 * @param args The initialized connection handler
 * @return void* This is a thread, returns NULL
 */
void* server_thread(void *args)
{
	Connection* conn = (Connection *) args;
	Request* req;
	char reply[conn->global_conf->max_reply_size + 1];

	// cycle forever until unrecoverable error happens
	while (! redir_should_exit) {
		// Connect to server if it is not connected yet or previous operation failed
		if (conn->state != ST_IDLE) {
			st_connect(conn);
		}

		// wait for next request and get it when it is ready
		req = redir_pop_request(conn);
		if (req == NULL) {
			continue;
		}

		conn->state = ST_BUSY;

		// try to send the request
		if (st_send(conn, req) < 0) {
			// if sending fails put the request back to the queue
			redir_dispatch_request(req, 1);
		} else {
			// try to receive reply
			if (st_receive(conn, reply) < 0) {
				// if receiveing fails put the request back to the queue
				redir_dispatch_request(req, 1);
			} else {
				// write reply to stdout
				redir_write_reply(req, reply, conn->global_conf);
				conn->state = ST_IDLE;
			}
		}
	}

	// close the connection
	close(conn->socket);
	conn->socket = -1;
	conn->state = ST_FATAL;

	return NULL;
}

/**
 * @brief Waits file descriptor for read or write
 *
 * It splits a long select timeout to series of short timeouts
 * to allow to check 'should exit' flag at least each second (TIMEOUT_STEP)
 *
 * @param type What event we wait for 0 = read, 1 = write
 * @param fd The file descriptor for read or write
 * @param timeout The timeout to wait
 * @return int The last select result
 */

static int st_wait(int type, int fd, int timeout)
{
	int ret;
	int cur_timeout = TIMEOUT_STEP;
	fd_set fset;
	struct timeval tv;
	int max_timeout = timeout + TIMEOUT_STEP;

	// we wait 'max_timeout' or if it is >TIMEOUT_STEP
	// wait a number of steps

	do {
		// prepare data for select
		FD_ZERO(&fset);
		FD_SET(fd, &fset);

		// wait the TIMEOUT_STEP or the rest of max_timeout if it's less than step
		if (timeout < cur_timeout) {
			int t = max_timeout - cur_timeout;
			tv.tv_sec = t / 1000;
			tv.tv_usec = (t % 1000) * 1000;
		} else {
			tv.tv_sec = TIMEOUT_STEP / 1000;
			tv.tv_usec = (TIMEOUT_STEP % 1000) * 1000;
		}

		if (type == 0) {
			// check if it is ready to read
			ret = select( fd + 1, &fset, NULL, NULL, &tv);
		} else {
			// check if it is ready to write
			ret = select( fd + 1, NULL, &fset, NULL, &tv);
		}

		cur_timeout += TIMEOUT_STEP;
	} while (ret == 0 && !redir_should_exit && cur_timeout < max_timeout);

	return ret;
}


/**
 * @brief The connection thread procedure
 *
 * @param conn The connection handler
 * @return int 0 on success, otherwise failure
 */
int st_connect(Connection* conn)
{
	int ret;
	struct addrinfo *result=NULL;

	// if opreation was interrupted because of timeout
	if (conn->state == ST_WRTIMEOUT) {		// 'send' was interrupted

		// try to wait until it become ready (instead of reopen the connection)

		ret = st_wait(1, conn->socket, conn->global_conf->max_timeout);

		if (redir_should_exit) {
			goto exit_thread;
		}

		if (ret > 0) { 		// it is ready to write
			// mark it as working and return
			conn->state = ST_IDLE;

			// increment the counter of "ready to serv" threads
			redir_st_up();
			return 0;

		} else { 	// Got an error or timeout expired
			log_send(LOG_WARNING, "Delayed write timeout expired for server host %s:%d\n", conn->hostname, conn->port);
			// close the socket
			close(conn->socket);
			conn->socket = -1;
			// mark this thread as diconnected
			conn->state = ST_DISCONNECTED;
		}

	} else if (conn->state == ST_RDTIMEOUT) {	// 'recv' was interrupted
		// try to wait until it become ready
		ret = st_wait(0, conn->socket, conn->global_conf->max_timeout);

		if (redir_should_exit) {
			goto exit_thread;
		}

		if (ret > 0) { 		// it is ready to read
			char buff[BUFFSIZE];
			// discard input because we already redispatched
			// the request that waited this reply
			recv(conn->socket, buff, BUFFSIZE, 0);

			// mark it as working and return
			conn->state = ST_IDLE;

			// increment the counter of "ready to serv" threads
			redir_st_up();
			return 0;

		} else { 	// Got an error or timeout expired
			log_send(LOG_WARNING, "Delayed read timeout expired for server %s:%d\n", conn->hostname, conn->port);

			// close the socket
			close(conn->socket);
			conn->socket = -1;
			// mark this thread as diconnected
			conn->state = ST_DISCONNECTED;
		}
	}

	// We are here if the state is "disconnected"
	// try to reconnect until success of unrecoverable error
	while (conn->state != ST_IDLE) {
		struct addrinfo hints;
		struct addrinfo *rp;
		int n;
		char port[sizeof("65535")];

		// if we don't have a socket yet create it now
		if (conn->socket < 0) {
			unsigned long fionbio = 1;

			// try to create a socket;
			conn->socket = socket(AF_INET, SOCK_STREAM, 0);

			// if could not
			if (conn->socket < 0) {
				// if it can't create a socket don't continue
				log_send(LOG_ERR, "Can not create socket\n");
				continue;
			}
			// set non-blocking mode
			ioctl (conn->socket, FIONBIO, &fionbio);
		}

		memset(&hints, 0, sizeof(struct addrinfo));
		hints.ai_family = AF_INET;
		hints.ai_socktype = SOCK_STREAM;
		sprintf(port, "%u", 0xffff & conn->port);

		if ( result != NULL ) {
			freeaddrinfo(result);
			result=NULL;
		}
		n = getaddrinfo(conn->hostname, port, &hints, &result);
		if (n != 0){
			log_send (LOG_ERR, "getaddrinfo failed %s:%s\n", conn->hostname, port);
			sleep(1);
			continue;
		}
		for (rp = result ; rp != NULL ; rp = rp->ai_next) {
			// try to connect
			ret = st_connect_nointr(conn->socket, rp->ai_addr, rp->ai_addrlen);

			// don't check the result code for success because for non-blocking socket
			// it can not be connected so soon

			if (ret && errno != EINPROGRESS) {
				char error_buff[256];
				// get the error description for logging
				strerror_r(errno, error_buff, 256);
				log_send(LOG_ERR, "Can not connect to server '%s': %s\n", conn->hostname, error_buff);
				continue;
			}

			// wait until it become 'ready to write'
			ret = st_wait(1, conn->socket, conn->global_conf->connect_timeout);

			if (redir_should_exit) {
				goto exit_thread;
			}

			if (ret > 0) { 		// it is ready
				int err;
				socklen_t err_len = sizeof(err);

				// try to get the connect result (or error) code
				if ( getsockopt(conn->socket, SOL_SOCKET, SO_ERROR, (void*) &err, &err_len) < 0) {
					// it means our internal error and should not happen in real work
					log_send(LOG_ERR, "Can not get socket option\n");

					close (conn->socket);
					conn->socket = -1;
					break;
				}

				if (err == 0) {		// no error
					// mark it as working
					conn->state = ST_IDLE;

					// increment the counter of "ready to serv" threads
					redir_st_up();

					// if it is the first attempt unlock the main thread
					if (conn->start_lock != NULL) {
						pthread_rwlock_unlock(conn->start_lock);
						conn->start_lock = NULL;
					}
					if ( result != NULL ) {
						freeaddrinfo(result);
						result=NULL;
					}
					return 0;

				} else {	// we have an error
					char error_buff[256];
					int t;

					// get the error description for logging
					strerror_r(err, error_buff, 256);

					// if it is the first attempt unlock the main thread
					if (conn->start_lock != NULL) {
						pthread_rwlock_unlock(conn->start_lock);
						conn->start_lock = NULL;
					}
					// if it can't connect because the server is down or unreachable
					// we can try to conenct later

					// log the error
					log_send(LOG_WARNING, "The server '%s' is down or unreachable: %s\n", conn->hostname, error_buff);

					// close the socket
					close(conn->socket);
					conn->socket = -1;
					// set the new state (it is disconnected and need to wait awhile)
					conn->state = ST_DISCONNECTED;

					// We don't want to flood network with our reconnect attempts
					// wait a reconnect delay before try again
					// (wait N "timeout steps")
					for (t = TIMEOUT_STEP; t <= conn->global_conf->reconnect_delay; t += TIMEOUT_STEP) {
						usleep(TIMEOUT_STEP * 1000);
						if (redir_should_exit) {
							goto exit_thread;
						}
					}
					// (wait the rest: delay % "timeout step")
					t = conn->global_conf->reconnect_delay % TIMEOUT_STEP;
					if (t > 0) {
						usleep(t * 1000);
						if (redir_should_exit) {
							goto exit_thread;
						}
					}
					continue;
				}
			} else { 	// Got a select error or timeout expired
				log_send(LOG_WARNING, "Connect timeout expired for server %s\n", conn->hostname);

				// close the socket
				close(conn->socket);
				conn->socket = -1;
				conn->state = ST_DISCONNECTED;

				// if it is the first attempt unlock the main thread
				if (conn->start_lock != NULL) {
					pthread_rwlock_unlock(conn->start_lock);
					conn->start_lock = NULL;
				}
			}
		} // end of connect retry cycle
	}

exit_thread:
	if ( result != NULL ) {
		freeaddrinfo(result);
		result=NULL;
	}

	// close the socket if it is open
	if (conn->socket > 0) {
		close(conn->socket);
		conn->socket = -1;
	}

	// mark the thread as stopped
	conn->state = ST_FATAL;
	redir_st_exit();

	// if it is the first attempt unlock the main thread
	if (conn->start_lock != NULL) {
		pthread_rwlock_unlock(conn->start_lock);
		conn->start_lock = NULL;
	}

	// and stop the thread
	pthread_exit(NULL);

	// we should never reach this point
	return -1;
}

/**
 * @brief Sent a request to server
 *
 * @param conn The connection handler
 * @param request The request to send
 * @return int 0 on success
 */
int st_send(Connection* conn, Request* request)
{
	char* ptr = request->message;
	int size = request->size;
	int sent;

	// variables for select()
	fd_set	wr;
	struct timeval tv;
	int ret;

	// repeat until everything is sent or get an error
	while (size > 0) {

		// prepare data for select (check for ready to send)
		FD_ZERO(&wr);
		FD_SET(conn->socket, &wr);

		tv.tv_sec = conn->global_conf->write_timeout / 1000;
		tv.tv_usec = (conn->global_conf->write_timeout % 1000) * 1000;

		// check if it is ready to send
		ret = select( conn->socket + 1, NULL, &wr, NULL, &tv);

		if ( ret < 0 ) {
			// Got an error
			goto error;

		} else if (ret == 0) {
			// timeout
			// mark this thread as having timeout
			conn->state = ST_WRTIMEOUT;
			// decrease the number of "ready to serv" threads
			redir_st_down();

			return -1;	// ERROR
		}

		// try to send
		sent = send(conn->socket, ptr, size, 0);
		if (sent <= 0) {
			// Got an error
			goto error;
		}
		// shift the pointer
		ptr += sent;
		size -= sent;
	}

	log_send(LOG_INFO, "Request sent: %d %s\n", request->size, request->message);

	return 0;	// SUCCESS

error:
	log_send(LOG_INFO, "Close socket on request send\n");
	// close the socket
	close(conn->socket);
	conn->socket = -1;
	// mark this thread as diconnected
	conn->state = ST_DISCONNECTED;
	// decrease the number of "ready to serv" threads
	redir_st_down();

	return -1;	// ERROR
}

/**
 * @brief Get a reply from server
 *
 * @param conn The connection handler
 * @param reply The reply buffer
 * @return int 0 on success, otherwise failure
 */
int st_receive(Connection* conn, char* reply)
{
	// variables for select()
	fd_set	rd;
	struct timeval tv;
	int ret;
	int size = 0;
	char *ptr = reply;

	// we read replies until get a newline char or buffer is full
	while (1) {

		// prepare data for select (check for ready to read)
		FD_ZERO(&rd);
		FD_SET(conn->socket, &rd);

		tv.tv_sec = conn->global_conf->read_timeout / 1000;
		tv.tv_usec = (conn->global_conf->read_timeout % 1000) * 1000;

		// check if it is ready to read
		ret = select( conn->socket + 1, &rd, NULL, NULL, &tv);

		if ( ret < 0 ) {
			// Got an error
			log_send(LOG_INFO, "Got select error on reply read");
			goto error;

		} else if (ret == 0) {
			// timeout
			log_send(LOG_INFO, "Got select timeout on reply read");
			// mark this thread as having timeout
			conn->state = ST_RDTIMEOUT;
			// decrease the number of "ready to serv" threads
			redir_st_down();

			return -1;	// ERROR
		}

		// try to receive
		ret = recv(conn->socket, ptr, conn->global_conf->max_reply_size - size, 0);

		// if connection closed or error
		if (ret <= 0) {
			// Got an error
			log_send(LOG_INFO, "Got read error on reply read");
			goto error;
		}
		// if we got a complete line
		if (ptr[ret - 1] == '\n') {
			break;
		}

		// count the total size and shift the buffer pointer
		size += ret;
		ptr += ret;

		// if the buffer is full
		if (size >= conn->global_conf->max_reply_size ) {
			reply[conn->global_conf->max_reply_size] = '\n';
			// ret needs to always be num bytes received
			// after 'ptr'
			ret = 0;
			break;
		}
	}
	ptr[ret] = '\0';

	log_send(LOG_INFO, "Reply received: %d %s\n", ptr - reply + ret, reply);

	return 0;	// SUCCESS

error:
	// close the socket
	log_send(LOG_INFO, "Close socket on reply read\n");
	close(conn->socket);
	conn->socket = -1;
	// mark this thread as diconnected
	conn->state = ST_DISCONNECTED;
	// decrease the number of "ready to serv" threads
	redir_st_down();

	return -1;	// ERROR
}

/**
 * @brief Make sure there haven't been any errors or disconnects on this connection
 *
 * If the other side has closed it or there's been some other error, this
 * function will close the connection and mark it as disconnected.
 *
 * @param conn the connection we're checking
 *
 * @return 0 if the connection is fine
 *        -1 if the socket has disconnected or has an error
 */
int st_check_connection(Connection *conn)
{
	char buff;
	int ret;

	// try to read one byte
	ret = recv(conn->socket, &buff, 1, MSG_PEEK);

	// check the result
	if (ret < 0 && errno == EAGAIN) {	// it's OK: nothing to read on non-blocking socket
		return 0;
	}
	// else: connection is closed or some error except EAGAIN

	// close the socket
	close(conn->socket);
	conn->socket = -1;
	// mark this thread as diconnected
	conn->state = ST_DISCONNECTED;
	// decrease the number of "ready to serv" threads
	redir_st_down();

	return -1;
}
